from pathlib import Path

import caikit
import pytest

from .fixtures.docker import *
from .fixtures.grpc import *
from .fixtures.http import *
from .fixtures.mocked_results import *
from .fixtures.tls import *
from .fixtures.utils import ConnectionType


def pytest_configure(config):
    pass


@pytest.fixture(scope="session")
def monkeysession():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture
def model_name(request: pytest.FixtureRequest):
    """name of the model utilized by the tests. Has to be in `tests/tiny_models`"""
    # Note that this can be overridden in tests via indirect parametrization
    if "caikit_tgis_service" in request.fixturenames:
        return "flan-t5-small-caikit"

    available_models = [
        "T5ForConditionalGeneration-caikit",
        "BertForSequenceClassification-caikit",
        "BloomForCausalLM-caikit",
    ]
    return available_models[0]


@pytest.fixture(
    autouse=True,
    scope="session",
    params=[ConnectionType.INSECURE, ConnectionType.TLS, ConnectionType.MTLS],
)
def connection_type(request):
    yield request.param


@pytest.fixture(scope="session")
def caikit_nlp_runtime(
    grpc_server_thread_port,  # noqa: F811
    http_server_thread_port,  # noqa: F811
    connection_type,
    server_key_file,
    server_cert_file,
    ca_cert_file,
):
    """configures caikit for local testing"""
    models_directory = str(Path(__file__).parent / "tiny_models")

    tgis_backend_config = [
        {
            "type": "TGIS",
            "config": {"connection": {"hostname": "localhost:8033"}},
        }
    ]
    config = {
        "merge_strategy": "merge",
        "runtime": {
            "metrics": {"enabled": False},
            "local_models_dir": models_directory,
            "library": "caikit_nlp",
            "lazy_load_local_models": True,
            "grpc": {"enabled": True, "port": grpc_server_thread_port},
            "http": {
                "enabled": True,
                "port": http_server_thread_port,
                "server_shutdown_grace_period_seconds": 2,
            },
        },
        "model_management": {
            "initializers": {
                "default": {
                    "type": "LOCAL",
                    "config": {"backend_priority": tgis_backend_config},
                }
            }
        },
        "log": {"formatter": "pretty"},
    }
    if connection_type is ConnectionType.TLS:
        config["runtime"]["tls"] = {
            "server": {
                "key": server_key_file,
                "cert": server_cert_file,
            }
        }

    if connection_type is ConnectionType.MTLS:
        config["runtime"]["tls"] = {
            "server": {
                "key": server_key_file,
                "cert": server_cert_file,
            },
            "client": {
                "cert": ca_cert_file,
            },
        }

    caikit.config.configure(config_dict=config)



from types import GeneratorType

import pytest
import requests
from caikit_nlp_client import HttpClient
from requests.exceptions import SSLError

from .conftest import ConnectionType


def test_generate_text(
    http_client, model_name, prompt, mocker, accept_self_signed_certs
):
    generated_text = http_client.generate_text(model_name, prompt)
    assert isinstance(generated_text, str)
    assert generated_text


def test_generate_text_with_optional_args(
    http_client,
    model_name,
    generated_text_result,
    prompt,
    mocker,
    accept_self_signed_certs,
):
    mock = mocker.spy(requests, "post")

    generated_text = http_client.generate_text(
        model_name, prompt, max_new_tokens=20, min_new_tokens=4
    )

    assert isinstance(generated_text, str)
    assert generated_text
    assert mock.call_args_list[-1].kwargs["json"]["parameters"]["max_new_tokens"] == 20
    assert mock.call_args_list[-1].kwargs["json"]["parameters"]["min_new_tokens"] == 4


def test_timeout_kwarg(
    http_client, model_name, prompt, mocker, accept_self_signed_certs
):
    mock = mocker.spy(requests, "post")

    http_client.generate_text(model_name, prompt)
    assert mock.call_args_list[-1].kwargs["timeout"] == 60.0

    http_client.generate_text(model_name, prompt, timeout=42.0)
    assert mock.call_args_list[-1].kwargs["timeout"] == 42.0


def test_generate_text_with_no_model_id(http_client):
    with pytest.raises(ValueError, match="request must have a model id"):
        http_client.generate_text("", "dummy")


def test_generate_text_stream(
    pytestconfig,
    http_client,
    model_name,
    generated_text_stream_result,
    prompt,
    accept_self_signed_certs,
):
    if not pytestconfig.option.real_caikit:
        pytest.skip(
            reason="stream mocking is broken, see https://github.com/opendatahub-io/caikit-nlp-client/issues/46"
        )

    response = http_client.generate_text_stream(
        model_name,
        prompt,
    )

    assert isinstance(response, GeneratorType)
    response_list = list(response)
    assert response_list
    assert all(isinstance(text, str) for text in response_list)


def test_generate_text_stream_with_optional_args(
    pytestconfig,
    http_client,
    model_name,
    generated_text_stream_result,
    prompt,
    accept_self_signed_certs,
    mocker,
):
    if not pytestconfig.option.real_caikit:
        pytest.skip(
            reason="stream mocking is broken, see https://github.com/opendatahub-io/caikit-nlp-client/issues/46"
        )

    mock = mocker.spy(requests, "post")
    response = http_client.generate_text_stream(
        model_name,
        prompt,
        preserve_input_text=False,
        max_new_tokens=20,
        min_new_tokens=4,
    )

    assert isinstance(response, GeneratorType)
    response_list = list(response)
    assert response_list
    assert all(isinstance(text, str) for text in response_list)
    assert mock.call_args_list[-1].kwargs["json"]["parameters"]["max_new_tokens"] == 20
    assert mock.call_args_list[-1].kwargs["json"]["parameters"]["min_new_tokens"] == 4


def test_get_text_generation_parameters(
    http_client, monkeysession, accept_self_signed_certs
):
    assert http_client.get_text_generation_parameters() == {
        "max_new_tokens": "integer",
        "min_new_tokens": "integer",
        "truncate_input_tokens": "integer",
        "decoding_method": "string",
        "top_k": "integer",
        "top_p": "number",
        "typical_p": "number",
        "temperature": "number",
        "repetition_penalty": "number",
        "max_time": "number",
        "exponential_decay_length_penalty": {
            "start_index": "integer",
            "decay_factor": "number",
        },
        "stop_sequences": "array",
        "seed": "integer",
        "preserve_input_text": "boolean",
    }


@pytest.mark.parametrize("connection_type", [ConnectionType.TLS], indirect=True)
def test_tls_enabled(
    model_name,
    http_server,
    monkeysession,
    ca_cert_file,
    connection_type,
):
    assert connection_type is ConnectionType.TLS, "TLS should be enabled for this test"

    http_client = HttpClient("https://{}:{}".format(*http_server))

    with pytest.raises(SSLError, match=".*CERTIFICATE_VERIFY_FAILED.*"):
        assert http_client.generate_text(model_name, "dummy text")

    # a valid certificate autority should validate the response with no extra args
    with monkeysession.context() as monkeypatch:
        monkeypatch.setenv("REQUESTS_CA_BUNDLE", ca_cert_file)

        assert http_client.generate_text(model_name, "dummy text")


@pytest.mark.parametrize("connection_type", [ConnectionType.MTLS], indirect=True)
def test_client_instantiation(
    ca_cert_file,
    client_key_file,
    client_cert_file,
    connection_type,
):
    # MTLS tests
    with pytest.raises(
        ValueError,
        match="Must provide both ca_cert_path, client_cert_path, client_key_path for "
        "mTLS",
    ):
        HttpClient(
            "https://localhost:8080",
            ca_cert_path=ca_cert_file,
            client_cert_path=client_cert_file,
        )
        HttpClient(
            "https://localhost:8080",
            ca_cert_path=ca_cert_file,
            client_key_path=client_key_file,
        )
        HttpClient(
            "https://localhost:8080",
            client_cert_path=client_cert_file,
            client_key_path=client_key_file,
        )



import socket
import time
from contextlib import closing
from enum import Enum
from typing import Callable, TypeVar

WAIT_TIME_OUT = 10


_T = TypeVar("_T")


def get_random_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        port = s.getsockname()[1]
        return port


class ConnectionType(Enum):
    INSECURE = 1
    TLS = 2
    MTLS = 3


def wait_until(
    pred: Callable[..., _T], timeout: float = WAIT_TIME_OUT, pause: float = 0.1
) -> _T:
    start = time.perf_counter()
    exc = None
    while (time.perf_counter() - start) < timeout:
        try:
            value = pred()
        except Exception as e:  # pylint: disable=broad-except
            exc = e
        else:
            return value
        time.sleep(pause)

    raise TimeoutError("timed out waiting") from exc



from types import GeneratorType

import pytest
from caikit_nlp_client.grpc_client import GrpcClient

from .fixtures.utils import ConnectionType


def test_generate_text(model_name, grpc_client, prompt):
    generated_text = grpc_client.generate_text(model_name, prompt)

    assert isinstance(generated_text, str)
    assert generated_text


def test_generate_text_with_optional_args(grpc_client, model_name, prompt):
    generated_text = grpc_client.generate_text(
        model_name,
        prompt,
        preserve_input_text=False,
        max_new_tokens=20,
        min_new_tokens=4,
    )

    assert isinstance(generated_text, str)
    assert generated_text
    # TODO: also validate passing of parameters


@pytest.mark.parametrize("connection_type", [ConnectionType.INSECURE], scope="session")
def test_context_manager(mocker, grpc_server, connection_type, request, model_name):
    with GrpcClient(*grpc_server, insecure=True) as client:
        close_spy = mocker.spy(client, "_close")
        assert client.generate_text(model_name, "dummy text")

        channel = client._channel
        channel_close_spy = mocker.spy(channel, "close")

    close_spy.assert_called()
    channel_close_spy.assert_called()


def test_generate_text_with_no_model_id(grpc_client):
    with pytest.raises(ValueError, match="request must have a model id"):
        grpc_client.generate_text("", "What does foobar mean?")


def test_generate_text_stream(
    model_name, grpc_client, generated_text_stream_result, prompt
):
    response = grpc_client.generate_text_stream(
        model_name,
        prompt,
    )

    assert isinstance(response, GeneratorType)
    response_list = list(response)
    assert response_list
    assert all(isinstance(text, str) for text in response_list)


def test_generate_text_stream_with_optional_args(
    model_name, grpc_client, generated_text_stream_result, prompt
):
    response = grpc_client.generate_text_stream(
        model_name,
        prompt,
        preserve_input_text=False,
        max_new_tokens=20,
        min_new_tokens=4,
    )

    assert isinstance(response, GeneratorType)
    response_list = list(response)
    assert response_list
    assert all(isinstance(text, str) for text in response_list)

    with pytest.raises(ValueError, match="Unsupported kwarg key='invalid_kwarg'"):
        grpc_client.generate_text(model_name, "dummy", invalid_kwarg=42)


def test_request_exception_handling(
    using_real_caikit,
    grpc_client,
    mock_text_generation,
    model_name,
):
    """force generation of an exception at text generation time to make
    sure the client returns useful information,"""
    exc_prefix = (
        "Exception raised during inference. This may be a problem with your input:"
    )
    stream_exc_prefix = "Exception iterating responses:"
    if using_real_caikit:
        prompt = "dummy"
        detail = "Value out of range: -1"
        match = f"{exc_prefix} {detail}"
        match_stream = f"{stream_exc_prefix} {detail}"
        kwargs = {
            # provide invalid kwargs
            "min_new_tokens": -1,
        }
    else:
        # mock_text_generation raises an exception when [[raise exception]] is in
        # the input text
        detail = "user requested an exception"
        prompt = "[[raise exception]] dummy"
        match = f"{exc_prefix} {detail}"
        match_stream = f"{stream_exc_prefix} {detail}"
        kwargs = {}

    with pytest.raises(
        RuntimeError,
        match=match,
    ):
        grpc_client.generate_text(
            model_name,
            prompt,
            **kwargs,
        )

    streaming_response = grpc_client.generate_text_stream(model_name, prompt, **kwargs)
    with pytest.raises(
        RuntimeError,
        match=match_stream,
    ):
        list(streaming_response)


def test_get_text_generation_parameters(grpc_client):
    assert grpc_client.get_text_generation_parameters() == {
        "text": "string",
        "max_new_tokens": "int64",
        "min_new_tokens": "int64",
        "truncate_input_tokens": "int64",
        "decoding_method": "string",
        "top_k": "int64",
        "top_p": "double",
        "typical_p": "double",
        "temperature": "double",
        "seed": "uint64",
        "repetition_penalty": "double",
        "max_time": "double",
        "exponential_decay_length_penalty": {
            "start_index": "int64",
            "decay_factor": "double",
        },
        "stop_sequences": "string",
        "preserve_input_text": "bool",
    }



import logging

import pytest
import requests

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def docker_compose_file(pytestconfig):
    return str(pytestconfig.rootdir / "tests/fixtures/resources/docker-compose.yml")


def pytest_addoption(parser):
    """Adds --real-caikit option to pytest to test vs a real caikit+tgis instance"""

    parser.addoption(
        "--real-caikit",
        action="store_true",
        default=False,
        help="Test against a real caikit+tgis instance",
    )


@pytest.fixture(scope="session")
def flan_t5_small_caikit(pytestconfig, monkeysession):
    """Points to a downloaded model that can be used with caikit.

    This may be downloaded if not available
    """
    model_dir = pytestconfig.rootdir / "tests/fixtures/resources/flan-t5-small-caikit"

    if not model_dir.isdir():
        model_dir.mkdir()

    if not (model_dir / "config.yml").exists():
        # Download the model so that it can be tested
        monkeysession.setenv("ALLOW_DOWNLOADS", 1)

        import caikit_nlp

        model = caikit_nlp.text_generation.TextGeneration.bootstrap(
            "google/flan-t5-small"
        )
        model.save(str(model_dir))

    return str(model_dir)


@pytest.fixture(scope="session")
def caikit_tgis_service(
    # NOTE: order is important: we need to make sure that the model is available before
    # requesting the `docker_services` fixture, which brings up the stack and mounts
    # volumes.  If model is not available, the stack will not come up.
    flan_t5_small_caikit,
    docker_services,
):
    """Spins up a caikit+tgis instance, returning a dict, grpc_port and http_port"""

    def is_up(url: str) -> False:
        try:
            return requests.get(url, timeout=5).status_code == 200
        except Exception:
            logger.debug(
                "Encountered exception while health checking caikit_tgis_service"
            )
            return False

    caikit_http_port = docker_services.port_for("caikit", 8080)
    caikit_grpc_port = docker_services.port_for("caikit", 8085)

    url_base = f"http://localhost:{caikit_http_port}"
    healthcheck_url = f"{url_base}/health"
    docker_services.wait_until_responsive(
        timeout=30.0, pause=0.1, check=lambda: is_up(healthcheck_url)
    )

    yield {
        "grpc_port": caikit_grpc_port,
        "http_port": caikit_http_port,
    }


@pytest.fixture(scope="session")
def grpc_server_docker(caikit_tgis_service):
    """Returns the grpc port for the caikit container"""
    return "localhost", caikit_tgis_service["grpc_port"]


@pytest.fixture(scope="session")
def http_server_docker(caikit_tgis_service):
    """Returns the http port for the caikit container"""
    return "localhost", caikit_tgis_service["http_port"]
